local socket = require 'socket'
local try, protect, newtry  = socket.try, socket.protect, socket.newtry
local setmetatable = setmetatable
local string, table, type  = string, table, type
local insert = table.insert
local tonumber, next = tonumber, next

module(...)

local MT -- metatable

local function cmd_respond(cln, cmd)
    try(cln.conn:send(cmd .. '\r\n'))
    local resp = try(cln.conn:receive())
    if string.sub(resp, 1, 3) == "+OK" then
        return resp
    else
        return nil, resp
    end
end

local function login(cln, user, pass, md5)
    local cmd
    local timestamp = string.match(cln.banner, "(<%S+>)")
    if md5 and timestamp then -- try APOP 
        local digest = md5(timestamp .. pass)
        cmd = string.format("APOP %s %s", user, digest)
        local resp, err = cmd_respond(cln, cmd)
        if resp then return resp end  -- authentication succeded
    end
    cmd= "USER " .. user
    try(cmd_respond(cln, cmd))
    cmd = "PASS " .. pass
    return cmd_respond(cln, cmd)
end

local function newclient(conf)
    local host = conf.host
    local user = conf.user
    local pass = conf.pass
    local port = conf.port or 110
    local create = conf.create or socket.tcp
    local md5 = conf.md5
    local cln = {}
    setmetatable(cln, MT)
    cln.conn = try(create())
    try(cln.conn:connect(host, port))
    local oldtry = try
    try = socket.newtry(function () cln.conn:close()end)
    cln.banner = try(cln.conn:receive())    -- save greeting line for APOP command
    try(login(cln, user, pass, md5))
    try = oldtry
    return cln
end

local function get_data_table(cln, fnc)
    fnc = fnc or insert
    local data = {}
    while true do
        local line = try(cln.conn:receive())
        if line == "." then break end
        -- unstuff the line
        if string.sub(line, 1, 1) == "." then line = string.sub(line, 2) end
        fnc(data, line)
    end
    return data
end

local function get_data(cln)
    local data = get_data_table(cln)
    insert(data, "")
    return table.concat(data, "\r\n")
end

local function stat(cln)
    local cmd = "STAT"
    local resp = try(cmd_respond(cln, cmd))
    return string.match(resp, "+OK (%d+) (%d+)")
end

local function quit(cln)
    local cmd = "QUIT"
    return cmd_respond(cln, cmd)
end

local function dele(cln, msg)
    local cmd = "DELE " .. msg
    return cmd_respond(cln, cmd)
end

local function rset(cln)
    local cmd = "RSET"
    return cmd_respond(cln, cmd)
end

local function noop(cln)
    local cmd = "NOOP"
    return cmd_respond(cln, cmd)
end

local function retr(cln, msg)
    local cmd = "RETR " .. msg
    try(cmd_respond(cln, cmd))
    return get_data(cln)
end

local function list(cln, msg)
    local cmd = "LIST"
    msg = tonumber(msg)
    if msg  then
        cmd = "LIST " .. msg
    end
    local resp = try(cmd_respond(cln, cmd))
    if msg then
        return string.match(resp, "+OK (%d+) (%d+)")
    else
        return get_data_table(cln, function (tab, line) 
                                       local k, v = string.match(line, "(%d+) (%d+)")
                                       tab[tonumber(k)] = v
                                   end)
    end
end

local function uidl(cln, msg)
    local cmd = "UIDL"
    msg = tonumber(msg)
    if msg  then
        cmd = "UIDL " .. msg
    end
    local resp = try(cmd_respond(cln, cmd))
    if msg  then
        return string.match(resp, "+OK (%d+) (%S+)")
    else
        return get_data_table(cln, function (tab, line) 
                                       local k, v = string.match(line, "(%d+) (%S+)")
                                       tab[tonumber(k)] = v
                                   end)
    end
end

local function top(cln, msg, lines)
    local cmd = string.format("TOP %d %d", msg, lines)
    try(cmd_respond(cln, cmd))
    return get_data(cln)
end

local function messages(cln)
    local lst = try(list(cln))
    local k = nil

    local function iter()
        k = next( lst, k)
        if not k then return nil end
        if not protect(list)(cln, k) then return iter() end
        local msg = try(retr(cln, k))
        return k, msg
    end
    return protect(iter)
end

local function close(cln)
    return cln.conn:close()
end

new = protect(newclient)

MT = {
--     login = protect(login),
    stat = protect(stat),
    quit = protect(quit),
    rset = protect(rset),
    dele = protect(dele),
    retr = protect(retr),
    list = protect(list),
    noop = protect(noop),
    top  = protect(top),
    uidl = protect(uidl),

    messages = protect(messages),
    close = close,
}
MT.__index = MT
